#!/usr/bin/perl

# Copyright 2003 by Benjamin Crowell. This software is copyleft licensed under
# the same terms as Perl, or, at your option, under the GPL 1.1 license.

use strict;
use locale;
use utf8;

#----------------------------------------------------------------
# defaults:

our %preferences=(
  'language'=>'en',                 # user's language
  'past'=>-1,                       # how many days into the past the report extends
  'future'=>14,                     # ...and how far into the future
  'calendar'=>'~/.when/calendar',   # where to find the calendar file
  'wrap'=>80,                       # 0 means don't wrap; otherwise, wrap display to this many columns
);

#----------------------------------------------------------------
# Strings are all collected here for ease of internationalization:

our $e_acute = "\x{e9}";

our %month_name =
  (
   'en'=>'Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec',   # English
   'fr'=>"Jan F${e_acute}v Mar Avr Mai Juin Juil Aug Sep Oct Nov D${e_acute}c", # French
  );
our %wday_name = ('en'=>'Sun Mon Tue Wed Thu Fri Sat');

sub w {
  my $what = shift;
  my @stuff = @_;
  my $lingo = $preferences{'language'};
  my %strings;
  my $found_lingo = 0;
  if (!$found_lingo) { # default to English
    $found_lingo = 1;
    %strings = 
      (
        'error_opening_prefs'=>"The preferences file %s exists, but there was an error opening it for input.",
        'syntex_err_in_prefs'=>"Syntax error in preferences file %s:\n%s",
        'prefs_file_not_found'=>"Preferences file %s not found.\n",
        'illegal_time_period'=>"Illegal time period, %s --- legal values are %s\n",
        'error_opening_calendar'=>"Couldn't open the file %s for input\n",
        'yesterday'=>'yesterday',
        'today'=>'today',
        'tomorrow'=>'tomorrow',
        'syntax_error_in_calendar'=>"Syntax error in calendar: %s\n",
        'illegal_year'=>"Illegal year: %s\n",
        'illegal_month'=>"Illegal month: %s\n",
        'illegal_day_of_month'=>"Illegal day of the month: %s\n",
      )
    }
  if (exists $strings{$what}) {$what = $strings{$what}}
  return sprintf($what,@stuff);
}


my $prefs_file = glob '~/.when/preferences';
if (-e $prefs_file) {
  open (FILE,"<$prefs_file") or die w('error_opening_prefs',$prefs_file);
  while (my $line = <FILE>) {
    if ($line =~ m/^\s*(\w+)\s*\=\s*([^\s]*)\s*/) {
      $preferences{lc($1)} = $2;
    }
    else {
      if (!($line =~ m/^\s*$/)) {die w('syntex_err_in_prefs',$prefs_file,$line)}
    }
  }
  close FILE;
}
else {
  print w('prefs_file_not_found',$prefs_file);
}
my $file = glob $preferences{'calendar'};


our $date_delimiter = ' '; # e.g. 2003 Feb 1
our $use_month_names = 1;
our $use_am_pm = 1;

our @month_length =      (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
#----------------------------------------------------------------

if (@ARGV) {
  my $period = lc($ARGV[0]);
  my %periods = ('w'=>7,'m'=>31,'y'=>366);
  if (!exists $periods{$period}) {die w('illegal_time_period',$period,(join ',',keys %periods))}
  $preferences{'future'} = $periods{$period};
}

my $now = When::current_date();

open (FILE, "<$file") or die w('error_opening_calendar',$file);
#binmode(FILE, ":utf8"); # bombs in Perl 5.6

print $preferences{'calendar'}."\n";

my @lines = <FILE>; # in array context, returns all lines from file

close FILE;

my @show;

foreach my $line(@lines) {
  chomp $line;
  if ($line =~ m/^\s*([^,#]*)\s*,\s*(([^\s].*)?)/) {
    my $what = $3;
    my $parsed_date = When::parse_blank_delimited($1);
    my ($err,$when) = @$parsed_date;
    if ($err ne '') {
      print "$err\n  $line\n";
    }
    if (!$err) {
      if ($when->y eq '*') {
        $when->y($now->y);
        if ($when->delta_days($now)<$preferences{'past'}) {$when->y($when->y+1)}
        if ($when->delta_days($now)>$preferences{'future'}) {$when->y($when->y-1)}
      }
      my $delta = $when->delta_days($now);
      if ($delta >= $preferences{'past'} && $delta <=$preferences{'future'}) {
        my $describe = $when->day_of_week_name;
        if ($delta == -1) {$describe = w('yesterday')}
        if ($delta == 0) {$describe = w('today')}
        if ($delta == 1) {$describe = w('tomorrow')}
        my $say = sprintf "%9s %s %s\n",$describe,$when->string_human,$what;
        push @show,[$when,$say];
      }
    }
  }
  else {
    if (!($line =~ m/^\s*$/ || $line =~ m/^#/)) {die w('syntax_error_in_calendar',$line)}
  }
}

@show = sort {$a->[0]->compare($b->[0])} @show;

my $wrap = $preferences{'wrap'};
if ($wrap==0) {$wrap=9999}
my $margin = 22; # This is the width of the column containing the date.
foreach my $thing(@show) {
  my ($when,$say) = @$thing;
  print format_item($say,$wrap,$margin);
}

# Format a line of text for output. Wrap long lines nicely.
sub format_item {
  my $text = shift;
  my $wrap = shift;
  my $margin = shift;
  chomp $text;
  my @stuff = split_a_line($text,$wrap,2);
  my $result = (shift @stuff)."\n";
  my @the_rest = ();
  if (@stuff) {
    @the_rest = split_a_line((shift @stuff),$wrap-$margin,9999);
  }
  foreach my $line(@the_rest) {
    $result = $result . (' ' x $margin) . "$line\n";
  }
  return $result;
}

# Split a long line into shorter pieces, preferably at word breaks.
sub split_a_line {
  my $text = shift;
  my $width = shift;
  my $max_pieces = shift;
  if (length($text)<=$width || $max_pieces==1) {return ($text,)}
  if ($text =~ m/^(.{0,$width})(\s+(.*))?$/) {
    return ($1,split_a_line($3,$width,$max_pieces-1))
  }
  # The only reason we'd get to this point is if they have an extremely long line
  # consisting of one extremely long word with no blanks in it. This means we can't
  # break at a word boundary.
  $text =~ m/^(.{0,$width})(.*)$/;
  return ($1,split_a_line($2,$width,$max_pieces-1));
}

#----------------------------------------------------------------

package When;

# Stores y, m, d, t.
# m is 1..12
# t is null, or hour, or hour:min; hour is on 24-hour time

sub new {
  my $class = shift;
  my $self = {};
  bless($self,$class);
  $self->{Y} = shift;
  $self->{M} = shift;
  $self->{D} = shift;
  if (@_) {
    $self->{T}=shift
  }
  else {
    $self->{T} = '';
  }
  return $self;
}

sub parse_blank_delimited {
  my $date = shift;
  chomp $date;
  my @a = split / /,$date;
  if ($a[0] ne '*' && $a[0]<1900 || $a[0]>2100) {return [main::w('illegal_year',$a[0])]}
  $a[1] = parse_month_name($a[1]);
  if ($a[1] eq '' || $a[1]<1 || $a[1]>12) {return [main::w('illegal_month',$a[1])]}
  if ($a[2]<1 || $a[2]>31) {return [main::w('illegal_day_of_month',$a[2])]}
  return ['',When->new(@a)];
}

# can be number or name
# if it's a name, ignores case, trailing dot, and extra characters after first 3 or 4
sub parse_month_name {
  my $name = shift;
  if ($name =~ m/\d+/) {return $name}
  for (my $m=1; $m<=12; $m++) {
    my $n = month_name($m);
    my $n2 = month_name($m,'en');
    if ($name =~ m/^$n/i ||  $name =~ m/^$n2/i) {return $m}
  }
  return '';
}

sub clone {
  my $self = shift;
  return When->new($self->array);
}

sub array {
  my $self = shift;
  return ($self->y,$self->m,$self->d,$self->t);
}

sub y {
  my $self = shift;
  if (@_) {$self->{Y} = shift}
  return $self->{Y};
}

sub m {
  my $self = shift;
  if (@_) {$self->{M} = shift}
  return $self->{M};
}

sub d {
  my $self = shift;
  if (@_) {$self->{D} = shift}
  return $self->{D};
}

sub t {
  my $self = shift;
  if (@_) {$self->{T} = shift}
  return $self->{T};
}

sub hour {
  my $self = shift;
  my $t = $self->t;
  $t =~ m/^\d+/;
  return $1;
}

# returns null string if not set
sub min {
  my $self = shift;
  my $t = $self->t;
  if ($t =~ m/\d+\:(\d_)/) {
    return $1;
  }
  else {
    return '';
  }
}

sub min_no_null {
  my $self = shift;
  my $m = $self->min;
  if ($m eq '') {$m=0}
  return $m;
}

sub current_date {
    my @tm = localtime;
    my $y = $tm[5];
    my $m = $tm[4]+1;
    my $d = $tm[3];
    if ($y<1900) {$y=$y+1900} # works in Perl 5 and 6
    return When->new($y,$m,$d);
}


sub string_sortable {
  my $self = shift;
  return sprintf "%04d-%02d-%02d %02d:%02d", $self->y,$self->m,$self->d,$self->hour,$self->min_no_null;
}

# y,m,d = numbers, m=1..12; t is h or h:m, on 24-hour time
sub string_human {
  my $when = shift;
  my ($y,$m,$d,$t) = $when->array;
  if ($use_month_names) {$m=month_name($m)}
  if ($d<10) {$d=" $d"}
  my $result = $y.$date_delimiter.$m.$date_delimiter.$d;
  if ($t ne '') {$result = $result . ' '.time_string_human($t)}
  return $result;
}

sub time_string_human {
  my $self = shift;
  my ($h,$m) = ($self->hour,$self->min);
  my $suffix = '';
  if ($use_am_pm) {
    if ($h>12) {
      $h=$h-12;
      $suffix = 'pm'
    }
    else {
      $suffix = 'am';
    }
  }
  if ($m ne '') {
    return sprintf '%d:%02d%s',$h,$m,$suffix;
  }
  else {
    return sprintf '%d%s',$h,$suffix;
  }
}

sub month_name {
  my $m = shift;
  my $language = $preferences{'language'};
  if (@_) {$language = shift}
  if ($m<1 || $m>12) {return ''}
  my $names = $month_name{'en'};
  if (exists $month_name{$language}) {
    $names = $month_name{$language}
  }
  my @names = split / /,$names;
  return $names[$m-1];
}

sub wday_name {
  my $d = shift;
  if ($d<1 || $d>7) {return ''}
  my $names = $wday_name{'en'};
  if (exists $wday_name{$preferences{'language'}}) {
    $names = $wday_name{$preferences{'language'}}
  }
  my @names = split / /,$names;
  return $names[$d-1];
}

sub compare {
  my $a = shift;
  my $b = shift;
  if ($a->y != $b->y) {return $a->y <=> $b->y}
  if ($a->m != $b->m) {return $a->m <=> $b->m}
  return $a->d <=> $b->d;
}

sub delta_days {
  my $a = shift;
  my $b = shift;
  my $compared = $a->compare($b);
  if ($compared == 0) {return 0}
  if ($compared == -1) {return -($b->delta_days($a))}
  if ($a->d != 1 || $b->d !=1) {
    my $aa = $a->clone;
    my $bb = $b->clone;
    $aa->{D} = 1;
    $bb->{D} = 1;
    return ($aa->delta_days($bb))+($a->d)-($b->d);
  }
  if ($a->m != 1 || $b->m !=1) {
    my $aa = $a->clone;
    my $bb = $b->clone;
    my $correction = 0;
    while ($aa->m > 1) {
      $aa->m($aa->m-1);
      $correction = $correction + $aa->days_in_month;
    }
    while ($bb->m > 1) {
      $bb->m($bb->m-1);
      $correction = $correction - $bb->days_in_month;
    }
    return ($aa->delta_days($bb))+$correction;
  }
  # From Jan 1 of one year to Jan 1 of another; $a is after $b
    my $aa = $a->clone;
    my $result = 0;
    while ($aa->y > $b->y) {
      $aa->y($aa->y-1);
      if ($aa->is_leap_year) {
        $result = $result+366;
      }
      else {
        $result = $result+365;
      }
    }
    return $result;
}

sub days_in_month {
  my $self = shift;
  if ($self->m != 2) {return $month_length[($self->m)-1]}
  if ($self->is_leap_year) {
    return 29;
  }
  else {
    return 28;
  }
}

sub is_leap_year {
  my $self = shift;
  my $y = $self->y;
  if ($y%4!=0) {return 0}
  if ($y%100!=0) {return 1}
  if ($y%1000!=0) {return 0}
  return 1;
}

sub day_of_week {
  my $self =shift;
  return ($self->delta_days(When->new(2003,2,2))%7)+1;
}

sub day_of_week_name {
  my $self =shift;
  return wday_name($self->day_of_week);
}

